// Copyright (c) 2005  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of cgal-python; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with cgal-python.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Id$
// $URL$
//
// Author(s)     : Spiros Evangelatos
//=========================================================================
//
#include <CGAL/basic.h>
#include <include/Kernels.h>
#include <include/iterator.h>
#include <CGAL/circulator.h>

#include <boost/python.hpp>
#include <boost/cstdint.hpp>

#include <CGAL/Polygon_2.h>
#include "pylist_helper.h"
#include "Polygon_2_docs.h"
#include <boost/shared_ptr.hpp>

using namespace boost::python;

//============Edge and Vertex iterators=============
template < class iterator, class Polygon>
simple_python_iterator<iterator> py_all_edges(Polygon& poly)
{
	std::pair< iterator , iterator > p( poly.edges_begin(), poly.edges_end() );
	return  simple_python_iterator<iterator>(p);
}
template < class iterator, class Polygon>
simple_python_iterator<iterator> py_all_vertices(Polygon& poly)
{
	std::pair< iterator , iterator > p( poly.vertices_begin(), poly.vertices_end() );
	return  simple_python_iterator<iterator>(p);
}
//============Edge and Vertex iterators end=============

//============Edge and Vertex circulators=============
template < class circulator, class Polygon>
simple_python_circulator<circulator> py_vertices_circulator(Polygon& poly)
{
	circulator first = poly.vertices_circulator();
	circulator last;
	std::pair< circulator , circulator > p( first, last );
	return  simple_python_circulator<circulator>(p);
}
template < class circulator, class Polygon>
simple_python_circulator<circulator> py_edges_circulator(Polygon& poly)
{
	circulator first = poly.edges_circulator();
	circulator last;
	std::pair< circulator , circulator > p( first, last );
	return  simple_python_circulator<circulator>(p);
}

//============Edge and Vertex circulators End=============

template<class Polygon_2>
str py_repr_Polygon_2(const Polygon_2& p )
{
	typedef typename Polygon_2::Vertex_iterator iterator;
	iterator i;
	if ( p.is_empty() ) {
		return str("Polygon_2()");
	}
	
	boost::python::str  s = "Polygon_2([";
	//append all but the last vertex a comma
	for (i = p.vertices_begin(); i != p.vertices_end() && (i + 1) != p.vertices_end(); ++i) {
		s = str( s + str(*i) + ", ");
	}
	// ommit the comma after the last vertex
	s = str( s + str(*i) );
	s = str(s + "])");
	return s;
}

template < class Point_2, class itterator, class Polygon>
void py_insert_at(Polygon& poly, int pos, Point_2& p)
{
	if ( pos < 0 || pos > poly.size()) {
		PyErr_SetString(PyExc_IndexError, "Index out of range");
		throw_error_already_set();
	}
	itterator at = poly.vertices_begin() + pos;
	poly.insert(at, p);
}

template <class Polygon>
void py_erase(Polygon & poly, int i)
{
	//support pythonic negative indexes
	//and raise index out of range exception
	i = sanitize_py_index(i, poly.size());
	poly.erase(poly.vertices_begin() + i);
}

template <class Polygon, class Point>
void py_set(Polygon & poly, int i, const Point &p)
{
	//support pythonic negative indexes
	//and raise index out of range exception
	i = sanitize_py_index(i, poly.size());
	poly.set(poly.vertices_begin() + i, p);
}

/* in order to protect from out of bounds access */
template <class Polygon, class Edge>
 Edge py_edge(Polygon & poly, int i)
{
	// Don't accept negative indices.
	check_positive_index(i, poly.size());
	return poly.edge(i);
}

//============ left/right/top/bottom vertex queries =============
template<class Polygon>
typename std::iterator_traits<typename Polygon::Vertex_iterator>::difference_type
py_left_vertex(Polygon & poly)
{
	return std::distance(poly.vertices_begin(), poly.left_vertex());
}

template<class Polygon>
typename std::iterator_traits<typename Polygon::Vertex_iterator>::difference_type
py_right_vertex(Polygon & poly)
{
	return std::distance(poly.vertices_begin(), poly.right_vertex());
}

template<class Polygon>
typename std::iterator_traits<typename Polygon::Vertex_iterator>::difference_type
py_top_vertex(Polygon & poly)
{
	return std::distance(poly.vertices_begin(), poly.top_vertex());
}

template<class Polygon>
typename std::iterator_traits<typename Polygon::Vertex_iterator>::difference_type
py_bottom_vertex(Polygon & poly)
{
	return std::distance(poly.vertices_begin(), poly.bottom_vertex());
}

//============ left/right/top/bottom vertex queries end =============

//============ Constructor from python list start =============
//Normally, in Boost 1.34 we would do something like that
//Unfortunatelly, Boost in debian and the most widely deployed boost libs
//don't support the stl_input_iterator conversion from a python iterable to 
//stl:iterators, so as a workaround, we use a python list.
/*
Polygon_2_from_list(const boost::python::object & o, const Traits & p_traits = Traits())
	: Polygon_2(p_traits)
{
	
	stl_input_iterator<Point_2> begin(o), end;
	insert( vertex_end(), begin, end);
}
*/

template<class Polygon, class Point>
boost::shared_ptr<Polygon>
Polygon_2_from_list(const boost::python::list & lst)
{
	boost::shared_ptr<Polygon> p (new Polygon);
	int i,n = extract<int>(lst.attr("__len__")());
	for ( i = 0; i<n; i++) {
		//The extraction of the C++ Point_2 object causes the warning:
		//dereferencing type-punned pointer will break strict-aliasing rules
		// The same warning occurs in many places across cgal-python. 
		p->push_back(extract<Point>(lst[i])); 
	}
	return p;
}

//============ Constructor from python list end =============
	
template<class kernel>
void Py_Polygon_2()
{
	typedef typename CGAL::Polygon_2<kernel>			Polygon_2; 
	typedef typename kernel::Point_2 					Point_2;
	typedef typename kernel::Segment_2 					Segment_2;
	typedef typename Polygon_2::Vertex_iterator			Vertex_iterator;
	typedef typename Polygon_2::Vertex_circulator	                Vertex_circulator;
	typedef typename Polygon_2::Edge_const_iterator		Edge_const_iterator;
	typedef typename Polygon_2::Edge_const_circulator	Edge_const_circulator;


	class_< Polygon_2 >("Polygon_2", Polygon_2_docs::polygon, init< optional< const kernel& >  >(""))
        .def(init< const Polygon_2& >(""))
	.def("__init__", make_constructor(&Polygon_2_from_list<Polygon_2, Point_2>) )
	
	.def("insert", & py_insert_at<Point_2, Vertex_iterator, Polygon_2>, Polygon_2_docs::insert)
	.def("push_back", & Polygon_2::push_back, Polygon_2_docs::push_back )
	.def("clear", & Polygon_2::clear, Polygon_2_docs::clear )
	.def("reverse_orientation", & Polygon_2::reverse_orientation, Polygon_2_docs::reverse_orientation )
	.def("erase", &py_erase<Polygon_2>, Polygon_2_docs::erase )
	.def("is_simple", & Polygon_2::is_simple, Polygon_2_docs::is_simple )
	.def("is_convex", & Polygon_2::is_convex, Polygon_2_docs::is_convex )
	.def("orientation", & Polygon_2::orientation, Polygon_2_docs::orientation )
	.def("oriented_side", & Polygon_2::oriented_side, Polygon_2_docs::oriented_side )
	.def("bounded_side", & Polygon_2::bounded_side, Polygon_2_docs::bounded_side )
	.def("bbox", & Polygon_2::bbox, Polygon_2_docs::bbox )
	.def("area", & Polygon_2::area, Polygon_2_docs::area )
	.def("left_vertex_index", &py_left_vertex<Polygon_2>, Polygon_2_docs::left_vertex )
	.def("right_vertex_index", &py_right_vertex<Polygon_2>, Polygon_2_docs::right_vertex )
	.def("top_vertex_index", &py_top_vertex<Polygon_2>, Polygon_2_docs::top_vertex )
	.def("bottom_vertex_index", &py_bottom_vertex<Polygon_2>, Polygon_2_docs::bottom_vertex )
	.def("is_counterclockwise_oriented", & Polygon_2::is_counterclockwise_oriented, 
		Polygon_2_docs::is_counterclockwise_oriented )
	.def("is_clockwise_oriented", & Polygon_2::is_clockwise_oriented, 
		Polygon_2_docs::is_clockwise_oriented )
	.def("is_collinear_oriented", & Polygon_2::is_collinear_oriented, 
		Polygon_2_docs::is_collinear_oriented )
	.def("has_on_positive_side", & Polygon_2::has_on_positive_side, 
		Polygon_2_docs::has_on_positive_side )
	.def("has_on_negative_side", & Polygon_2::has_on_negative_side, 
		Polygon_2_docs::has_on_negative_side )
	.def("has_on_boundary", & Polygon_2::has_on_boundary, 
		Polygon_2_docs::has_on_boundary )
	.def("has_on_bounded_side", & Polygon_2::has_on_bounded_side, 
		Polygon_2_docs::has_on_bounded_side )
	.def("has_on_unbounded_side", & Polygon_2::has_on_unbounded_side,
		Polygon_2_docs::has_on_unbounded_side )

	// implement python list-like interface
	// use with_custodian_and_ward to let container keep value
	.def("append", &pylist_helper<Polygon_2, Point_2>::add, with_custodian_and_ward<1,2>(), 
		Polygon_2_docs::append)
	.def("__getitem__", &pylist_helper<Polygon_2, Point_2>::get, return_value_policy<copy_const_reference>(),
		Polygon_2_docs::__getitem__) 
	.def("__setitem__", &py_set<Polygon_2, Point_2>, with_custodian_and_ward<1,2>(),
		Polygon_2_docs::__setitem__) 
	.def("__delitem__", &py_erase<Polygon_2>, Polygon_2_docs::__delitem__) 
	.def("__len__", & Polygon_2::size, Polygon_2_docs::__len__) 

	.def("size", & Polygon_2::size, Polygon_2_docs::size) 
	.def("is_empty", & Polygon_2::is_empty, Polygon_2_docs::is_empty) 
	.def("edge", &py_edge<Polygon_2, Segment_2>, Polygon_2_docs::edge) 
	.def("__repr__", &py_repr_Polygon_2<Polygon_2>, Polygon_2_docs::__repr__) 
	.def("__str__", &py_repr_Polygon_2<Polygon_2>, Polygon_2_docs::__str__) 
	
	.def( self == self ) 
	.def( self != self ) 
	.add_property("vertices", &py_all_vertices<Vertex_iterator, Polygon_2>, Polygon_2_docs::vertices)
	.add_property("edges", &py_all_edges<Edge_const_iterator, Polygon_2>, Polygon_2_docs::edges)
	.add_property("vertices_circulator", &py_vertices_circulator<Vertex_circulator, Polygon_2>, Polygon_2_docs::vertices_circulator)
	.add_property("edges_circulator", &py_edges_circulator<Edge_const_circulator, Polygon_2>, Polygon_2_docs::edges_circulator)
	;

    simple_python_iterator<Vertex_iterator>::declare("Vertex_iterator","\n\niterator over vertices");
    simple_python_circulator<Vertex_circulator>::declare("Vertex_circulator","\n\ncirculator over vertices");
    simple_python_iterator<Edge_const_iterator>::declare("Edge_iterator","\n\niterator over edges");
    simple_python_circulator<Edge_const_circulator>::declare("Edge_circulator","\n\ncirculator over edges");
}

void export_Polygon_2()
{
	Py_Polygon_2<K>();
}
